from collections.abc import Awaitable, Callable

import cv2
import numpy as np
import PIL

from ... import rosys
from ...geometry import Point
from ..image import ImageSize


class SimulatedDevice:

    def __init__(self,
                 id: str,  # pylint: disable=redefined-builtin
                 *,
                 size: ImageSize,
                 on_new_image_data: Callable[[bytes, float], Awaitable | None],
                 color: str = '#ffffff',
                 fps: float = 30.0) -> None:
        self._id = id
        self._size = size
        self.color = color
        self._on_new_image_data = on_new_image_data
        self._creation_time: float = rosys.time()

        self._repeater = rosys.on_repeat(self._create_image, interval=1.0 / fps)

    @property
    def creation_time(self) -> float:
        return self._creation_time

    async def _create_image(self) -> None:
        timestamp = rosys.time()
        image_data: bytes | None
        if rosys.is_test:
            image_data = b'test data'
        else:
            image_data = await rosys.run.cpu_bound(_create_image_data, self._id, self._size, self.color, timestamp)
        if not image_data:
            return
        result = self._on_new_image_data(image_data, timestamp)
        if isinstance(result, Awaitable):
            await result

    def set_fps(self, fps: float) -> None:
        self._repeater.interval = 1.0 / fps

    def get_fps(self) -> float:
        return 1.0 / self._repeater.interval


def _create_image_data(id: str, size: ImageSize, color: str, timestamp: float) -> bytes:  # pylint: disable=redefined-builtin
    img = PIL.Image.new('RGB', size=(size.width, size.height), color=color)
    d = PIL.ImageDraw.Draw(img)
    text = f'{id}: {timestamp:.2f}'
    position = _floating_text_position(size.width, size.height, timestamp)

    d.text((position.x, position.y), text, fill=(0, 0, 0))
    d.text((position.x + 1, position.y + 1), text, fill=(255, 255, 255))

    _, encoded_image = cv2.imencode('.jpg', np.array(img)[:, :, ::-1])  # NOTE: cv2 expects BGR
    return encoded_image.tobytes()


def _floating_text_position(box_width: int, box_height: int, timestamp: float, speed: float = 100, angle: float = np.deg2rad(45)) -> Point:
    """Calculate the position of the text that floats around the image."""
    return Point(x=(speed * timestamp * np.cos(angle)) % box_width,
                 y=(speed * timestamp * np.sin(angle)) % box_height)



from __future__ import annotations

import math
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from .object3d import Object3d
from .point import Point

if TYPE_CHECKING:
    from .pose3d import Pose3d


@dataclass(slots=True, kw_only=True)
class Point3d(Object3d):
    x: float = 0
    y: float = 0
    z: float = 0

    @classmethod
    def zero(cls) -> Point3d:
        return Point3d(x=0, y=0, z=0)

    @staticmethod
    def from_tuple(t: Sequence[float]) -> Point3d:
        """Create a Point3d from the three first elements of a sequence."""
        return Point3d(x=t[0], y=t[1], z=t[2])

    @staticmethod
    def from_point(p: Point, z: float = 0) -> Point3d:
        """Create a Point3d from a Point."""
        return Point3d(x=p.x, y=p.y, z=z)

    @property
    def tuple(self) -> tuple[float, float, float]:
        return (self.x, self.y, self.z)

    @property
    def array(self) -> np.ndarray:
        return np.array(self.tuple).reshape(3, 1)

    def distance(self, other: Point3d) -> float:
        return math.sqrt((other.x - self.x)**2 + (other.y - self.y)**2 + (other.z - self.z)**2)

    def projection(self) -> Point:
        return Point(x=self.x, y=self.y)

    def __str__(self) -> str:
        return f'Point3d({self.x:.3f}, {self.y:.3f}, {self.z:.3f})'

    def __add__(self, other: Point3d) -> Point3d:
        return Point3d(x=self.x + other.x, y=self.y + other.y, z=self.z + other.z)

    def __sub__(self, other: Point3d) -> Point3d:
        return Point3d(x=self.x - other.x, y=self.y - other.y, z=self.z - other.z)

    def transform_with(self, pose: Pose3d) -> Point3d:
        """Transform this pose with another pose."""
        return Point3d.from_tuple((np.dot(pose.rotation.R, self.array) + pose.translation_vector).flatten())



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import overload

import cv2
import numpy as np

from ..geometry import Frame3d, Point, Point3d, Pose3d, Rotation
from .image import Image, ImageSize


class CameraModel(str, Enum):
    PINHOLE = 'pinhole'
    FISHEYE = 'fisheye'
    OMNIDIRECTIONAL = 'omnidirectional'


@dataclass(slots=True, kw_only=True)
class OmnidirParameters:
    """The parameters for an omnidirectional camera.

    :param xi: The omnidirectional camera parameter xi.
    :param rotation: Rotation used to determine crop when undistorting to pinhole image.
    """
    xi: float = 0.0
    rotation: Rotation = field(default_factory=Rotation.zero)


@dataclass(slots=True, kw_only=True)
class Intrinsics:
    """The intrinsic parameters of a camera.

    :param model: The camera model to use.
    :param matrix: The camera matrix K.
    :param distortion: The distortion coefficients D.
    :param xi: The omnidirectional camera parameter xi (only for ``CameraModel.OMNIDIRECTIONAL``).
    :param rotation: An inner rotation matrix, useful for visual-inertial calibration or omnidirectional projection.
    :param size: The size of the image.
    """
    model: CameraModel = CameraModel.PINHOLE
    matrix: list[list[float]]
    distortion: list[float]
    omnidir_params: OmnidirParameters | None = None
    size: ImageSize

    @staticmethod
    def create_default(width: int = 800, height: int = 600, *, focal_length: float = 570) -> Intrinsics:
        size = ImageSize(width=width, height=height)
        K: list[list[float]] = [[focal_length, 0, size.width / 2],
                                [0, focal_length, size.height / 2],
                                [0, 0, 1]]
        D: list[float] = [0] * 5
        return Intrinsics(matrix=K, distortion=D, size=size)


log = logging.getLogger('rosys.world.calibration')


@dataclass(slots=True, kw_only=True)
class Calibration:
    """Represents the full calibration of a camera."""
    intrinsics: Intrinsics
    extrinsics: Pose3d = field(default_factory=Pose3d)

    @staticmethod
    def from_points(world_points: list[Point3d] | list[list[Point3d]],
                    image_points: list[Point] | list[list[Point]],
                    *,
                    image_size: ImageSize,
                    f0: float,
                    rational_model: bool = False,
                    camera_model: CameraModel = CameraModel.PINHOLE,
                    frame: Frame3d | None = None) -> Calibration:
        """Estimate the camera calibration from corresponding world and image points.

        :param world_points: The observed points in 3D world coordinates.
        :param image_points: The observed points in 2D image coordinates.
        :param image_size: The size of the image.
        :param f0: An initial guess for the focal length.
        :param rational_model: Whether to use the rational camera model (only applies to pinhole cameras).
        :param camera_model: The camera model to use.
        :param frame: The coordinate frame of the world points and the extrinsic camera pose.

        :return: The estimated camera calibration.
        """
        if rational_model and camera_model != CameraModel.PINHOLE:
            raise ValueError('Rational model is only supported for pinhole cameras')

        if isinstance(world_points[0], Point3d) or isinstance(image_points[0], Point):
            assert len(world_points) == len(image_points), 'Image and world points are not formatted equally'
            world_points = [world_points]  # type: ignore
            image_points = [image_points]  # type: ignore

        world_point_array = [np.array([p.tuple for p in view], dtype=np.float32).reshape((-1, 1, 3))  # type: ignore
                             for view in world_points]
        image_point_array = [np.array([p.tuple for p in view], dtype=np.float32).reshape((-1, 1, 2))  # type: ignore
                             for view in image_points]
        K0 = np.array([[f0, 0, image_size.width / 2], [0, f0, image_size.height / 2], [0, 0, 1]], dtype=np.float32)
        D0 = np.array([0.1, 0.4, -0.5, 0.2], dtype=np.float32).reshape(1, 4)

        xi: float = 0.0

        optimization_criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 200, 1e-6)

        if camera_model == CameraModel.PINHOLE:
            flags = cv2.CALIB_USE_INTRINSIC_GUESS
            if rational_model:
                flags |= cv2.CALIB_RATIONAL_MODEL
            _rms, K, D, rvecs, tvecs = cv2.calibrateCamera(
                objectPoints=world_point_array,
                imagePoints=image_point_array,
                imageSize=image_size.tuple,
                cameraMatrix=K0,
                distCoeffs=np.zeros((1, 5), dtype=np.float32),
                flags=flags,
                criteria=optimization_criteria,
            )
        elif camera_model == CameraModel.FISHEYE:
            flags = cv2.fisheye.CALIB_RECOMPUTE_EXTRINSIC
            flags |= cv2.fisheye.CALIB_CHECK_COND
            flags |= cv2.fisheye.CALIB_FIX_SKEW
            flags |= cv2.fisheye.CALIB_USE_INTRINSIC_GUESS
            _rms, K, D, rvecs, tvecs = cv2.fisheye.calibrate(  # pylint: disable=unpacking-non-sequence
                objectPoints=world_point_array,
                imagePoints=image_point_array,
                image_size=image_size.tuple,
                K=K0,
                D=D0,
                flags=flags,
                criteria=optimization_criteria,
            )
            if D.size != 4:
                raise ValueError(f'Fisheye calibration failed (got invalid distortion coefficients D="{D}")')
        elif camera_model == CameraModel.OMNIDIRECTIONAL:
            flags = cv2.omnidir.CALIB_USE_GUESS
            flags |= cv2.omnidir.CALIB_FIX_SKEW
            _rms, K, xi_arr, D, rvecs, tvecs, _status = cv2.omnidir.calibrate(  # pylint: disable=unpacking-non-sequence
                objectPoints=world_point_array,
                imagePoints=image_point_array,
                size=image_size.tuple,
                K=K0,
                xi=np.array([xi], dtype=np.float32),
                D=D0,
                flags=flags,
                criteria=optimization_criteria,
            )
            xi = xi_arr[0]
            if D.size != 4:
                raise ValueError(f'Omnidirectional calibration failed (got invalid distortion coefficients D="{D}")')
        else:
            raise ValueError(f'Unknown camera model "{camera_model}"')

        intrinsics = Intrinsics(matrix=K.tolist(),
                                distortion=D.tolist()[0],
                                size=image_size,
                                model=camera_model,
                                omnidir_params=OmnidirParameters(xi=xi))

        rotation = Rotation.from_rvec(rvecs[0]).T
        translation = -np.array(rotation.R).dot(tvecs[0])
        extrinsics = Pose3d(x=translation[0, 0],
                            y=translation[1, 0],
                            z=translation[2, 0],
                            rotation=rotation).in_frame(frame)

        return Calibration(intrinsics=intrinsics, extrinsics=extrinsics)

    @overload
    def project_to_image(self, point: Point3d) -> Point | None:
        """Project a 3D point to the image plane.

        :param point: The point to project.
        """

    @overload
    def project_to_image(self, coordinates: np.ndarray, *, frame: Frame3d | None = None) -> np.ndarray:
        """Project a 3D coordinate array to the image plane.

        :param coordinates: The 3D coordinates of the points to project.
        :param frame: The coordinate frame of the coordinates.
        """

    def project_to_image(self, *args, **kwargs) -> Point | np.ndarray | None:
        if isinstance(args[0], Point3d):
            point: Point3d = args[0]
            world_array = np.array([point.resolve().tuple], dtype=np.float64)
            image_array = self.project_to_image(world_array)
            if np.isnan(image_array).any():
                return None
            return Point(x=image_array[0, 0], y=image_array[0, 1])  # pylint: disable=unsubscriptable-object

        coordinates: np.ndarray = args[0]
        frame: Frame3d | None = kwargs.get('frame')
        world_extrinsics = self.extrinsics.relative_to(frame)
        R = world_extrinsics.rotation.matrix.astype(np.float64)
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ world_extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion, dtype=float)

        # pylint: disable=unpacking-non-sequence
        if self.intrinsics.model == CameraModel.PINHOLE:
            image_array, _ = cv2.projectPoints(coordinates, Rod, t, K, D)
        elif self.intrinsics.model == CameraModel.FISHEYE:
            image_array, _ = cv2.fisheye.projectPoints(coordinates.reshape(-1, 1, 3), Rod, t, K, D)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            assert self.intrinsics.omnidir_params is not None, 'Omnidirectional parameters are unset'
            xi = self.intrinsics.omnidir_params.xi
            image_array, _ = cv2.omnidir.projectPoints(coordinates.reshape(-1, 1, 3), Rod, t, K, xi, D)
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')
        # pylint: enable=unpacking-non-sequence

        if self.intrinsics.model != CameraModel.OMNIDIRECTIONAL:
            coordinates = coordinates.reshape(-1, 3)
            local_coordinates = (coordinates - world_extrinsics.translation) @ R
            image_array[local_coordinates[:, 2] < 0, :] = np.nan

        return image_array.reshape(-1, 2)

    @overload
    def project_from_image(self, image_coordinates: Point, target_height: float = 0) -> Point3d | None: ...

    @overload
    def project_from_image(self, image_coordinates: np.ndarray, target_height: float = 0) -> np.ndarray: ...

    def project_from_image(self, image_coordinates: Point | np.ndarray, target_height: float = 0) -> Point3d | None | np.ndarray:
        """Project a point in image coordinates to a plane in world xy dimensions at a given height.

        :param image_coordinates: The image coordinates to project.
        :param target_height: The height of the plane in world coordinates.

        :return: The world coordinates of the projected point.
        """
        if isinstance(image_coordinates, Point):
            image_points = np.array(image_coordinates.tuple, dtype=np.float32)
            world_points = self.project_from_image(image_points, target_height=target_height)
            if np.isnan(world_points).any():
                return None
            return Point3d(x=world_points[0, 0], y=world_points[0, 1], z=world_points[0, 2])  # pylint: disable=unsubscriptable-object

        world_extrinsics = self.extrinsics.resolve()
        image_rays = self.points_to_rays(image_coordinates.astype(np.float32).reshape(-1, 1, 2))
        objPoints = image_rays @ world_extrinsics.rotation.matrix.T
        Z = world_extrinsics.z
        t = world_extrinsics.translation_vector
        world_points = t.T - objPoints * (Z - target_height) / objPoints[:, 2:]

        reprojection = self.project_to_image(world_points)
        sign = objPoints[:, -1] * np.sign(Z)
        distance = np.linalg.norm(reprojection - image_coordinates.reshape(-1, 2), axis=1)
        world_points[np.logical_not(np.logical_and(sign < 0, distance < 2)), :] = np.nan

        return world_points

    def points_to_rays(self, image_points: np.ndarray) -> np.ndarray:
        """Convert image points to rays in homogeneous coordinates with respect to the camera coordinate frame."""
        K = np.array(self.intrinsics.matrix, dtype=np.float32).reshape((3, 3))
        D = np.array(self.intrinsics.distortion)
        if self.intrinsics.model == CameraModel.PINHOLE:
            undistorted = cv2.undistortPoints(image_points, K, D)
        elif self.intrinsics.model == CameraModel.FISHEYE:
            undistorted = cv2.fisheye.undistortPoints(image_points, K, D)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            assert self.intrinsics.omnidir_params is not None, 'Omnidirectional parameters are unset'
            R = self.intrinsics.omnidir_params.rotation.matrix.astype(np.float32)
            xi = np.array(self.intrinsics.omnidir_params.xi, dtype=np.float32)
            undistorted = cv2.omnidir.undistortPoints(image_points, K, D, xi=xi, R=R)
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

        return cv2.convertPointsToHomogeneous(undistorted).reshape(-1, 3)

    def undistort_points(self, image_points: np.ndarray, crop: bool = False) -> np.ndarray:
        """Generalized wrapper for undistorting image points.

        :param image_points: The image points to undistort.
        :param crop: Whether cropping is applied to the image during undistortion.

        :return: The undistorted image points.
        """
        K = np.array(self.intrinsics.matrix, dtype=np.float32).reshape((3, 3))
        D = np.array(self.intrinsics.distortion)
        if self.intrinsics.model == CameraModel.PINHOLE:
            new_K = self.get_undistorted_camera_matrix(crop=crop)
            return cv2.undistortPoints(image_points, K, D, P=new_K, R=np.eye(3))
        elif self.intrinsics.model == CameraModel.FISHEYE:
            new_K = self.get_undistorted_camera_matrix(crop=crop)
            return cv2.fisheye.undistortPoints(image_points, K, D, P=new_K)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            assert self.intrinsics.omnidir_params is not None, 'Omnidirectional parameters are unset'
            R = np.array(self.intrinsics.omnidir_params.rotation, dtype=np.float32)
            xi = np.array(self.intrinsics.omnidir_params.xi, dtype=np.float32)
            return cv2.omnidir.undistortPoints(image_points, K, D, xi=xi, R=R)
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

    def distort_points(self, image_points: np.ndarray, crop: bool = False) -> np.ndarray:
        """Generalized wrapper for distorting image points.

        :param image_points: The image points to distort.
        :param crop: Whether cropping is applied to the image during distortion.

        :return: The distorted image points.
        """
        K = np.array(self.intrinsics.matrix, dtype=np.float32).reshape((3, 3))
        D = np.array(self.intrinsics.distortion)

        if self.intrinsics.model == CameraModel.PINHOLE:
            new_K = self.get_undistorted_camera_matrix(crop=crop)
            return cv2.undistortPoints(image_points, K, D, P=new_K, R=np.eye(3))
        elif self.intrinsics.model == CameraModel.FISHEYE:
            new_K = self.get_undistorted_camera_matrix(crop=crop)
            normalized_points = np.linalg.inv(new_K) @ cv2.convertPointsToHomogeneous(image_points).reshape(-1, 3).T
            normalized_points = cv2.convertPointsFromHomogeneous(normalized_points.T).reshape(-1, 1, 2)
            return cv2.fisheye.distortPoints(normalized_points, K, D)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            raise NotImplementedError('Re-distortion for omnidirectional cameras is not supported')
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

    def get_undistorted_camera_matrix(self, crop: bool = False) -> np.ndarray:
        """Compute the camera matrix for the undistorted image.

        :param crop: Whether cropping is applied to the image during undistortion.

        :return: The camera matrix for the undistorted image.
        """
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        w, h = self.intrinsics.size.width, self.intrinsics.size.height
        if self.intrinsics.model == CameraModel.PINHOLE:
            new_K, _ = cv2.getOptimalNewCameraMatrix(K, D, (w, h), 1, (w, h), centerPrincipalPoint=True)
        elif self.intrinsics.model == CameraModel.FISHEYE:
            new_K = cv2.fisheye.estimateNewCameraMatrixForUndistortRectify(K, D, (w, h), np.eye(3),
                                                                           balance=0.0 if crop else 1.0,
                                                                           new_size=(w, h),
                                                                           fov_scale=1)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            if crop:
                logging.warning('Cropping is not yet supported for omnidirectional cameras')
            new_K = np.array([[w / 4, 0, w / 2],
                              [0, h / 4, h / 2],
                              [0, 0, 1]])
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

        return new_K

    def undistort_array(self, image_array: np.ndarray, crop: bool = False) -> np.ndarray:
        """Undistort an image represented as a numpy array.

        The image is expected to be decoded (in particular not encoded bytes of a JPEG image).

        :param image_array: The image to undistort.
        :param crop: Whether cropping is applied to the image during undistortion.

        :return: The undistorted image.
        """
        if image_array.shape[0] != self.intrinsics.size.height or image_array.shape[1] != self.intrinsics.size.width:
            log.warning('Image size does not match calibration size (image: %s, calibration: %s)',
                        image_array.shape, self.intrinsics.size)
            return image_array

        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        h, w = image_array.shape[:2]
        if self.intrinsics.model == CameraModel.PINHOLE:
            new_K, roi = cv2.getOptimalNewCameraMatrix(K, D, (w, h), 1, (w, h), centerPrincipalPoint=True)

            dst = cv2.undistort(image_array, K, D, None, new_K)
            if crop:
                x, y, roi_w, roi_h = roi
                dst = dst[y:y+roi_h, x:x+roi_w]
        elif self.intrinsics.model == CameraModel.FISHEYE:
            balance = 0.0 if crop else 1.0
            new_K = cv2.fisheye.estimateNewCameraMatrixForUndistortRectify(K, D, (w, h), np.eye(3), balance=balance)

            map1, map2 = cv2.fisheye.initUndistortRectifyMap(  # pylint: disable=unpacking-non-sequence
                K, D, np.eye(3), new_K, (w, h), cv2.CV_16SC2)
            dst = cv2.remap(image_array, map1, map2, interpolation=cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT)
        elif self.intrinsics.model == CameraModel.OMNIDIRECTIONAL:
            flags = cv2.omnidir.RECTIFY_PERSPECTIVE
            new_K = self.get_undistorted_camera_matrix(crop=crop)

            assert self.intrinsics.omnidir_params is not None, 'Omnidirectional parameters are unset'
            R = np.array(self.intrinsics.omnidir_params.rotation.R)
            xi = np.array(self.intrinsics.omnidir_params.xi)
            dst = cv2.omnidir.undistortImage(image_array, K=K, D=D, xi=xi, Knew=new_K,
                                             flags=flags, new_size=(w, h), R=R)
        else:
            raise ValueError(f'Unknown camera model "{self.intrinsics.model}"')

        return dst

    def undistort_image(self, image: Image) -> Image:
        """Undistort an image represented as an Image object.

        If you already have the image as an unencoded numpy array, use ``undistort_array`` instead.

        :param image: The image to undistort.

        :return: The undistorted image.
        """
        return Image(
            camera_id=image.camera_id,
            size=image.size,
            time=image.time,
            data=cv2.imencode('.jpg', self.undistort_array(image.to_array()))[1].tobytes(),
            is_broken=image.is_broken,
            tags=image.tags,
        )

    def get_undistorted_size(self) -> ImageSize:
        """Compute the size of the undistorted image after cropping.

        :return: The size of the undistorted image.
        """
        if self.intrinsics.model == CameraModel.PINHOLE:
            K = np.array(self.intrinsics.matrix)
            D = np.array(self.intrinsics.distortion)
            h, w = self.intrinsics.size.height, self.intrinsics.size.width
            _, roi = cv2.getOptimalNewCameraMatrix(K, D, (w, h), 1, (w, h), centerPrincipalPoint=True)
            return ImageSize(width=roi[2], height=roi[3])
        else:
            return self.intrinsics.size



import numpy as np
from nicegui import ui

from .camera import CalibratableCamera


class CameraSceneObject(ui.scene.group):

    def __init__(self,
                 camera: CalibratableCamera, *,
                 color: str = '#333333',
                 px_per_m: float = 1000,
                 show_name: bool = True,
                 show_axes: bool = True,
                 ) -> None:
        super().__init__()
        with self:
            with ui.scene.group() as pyramid:
                ui.scene.cylinder(0, np.sqrt(0.5), 1, 4, wireframe=True) \
                    .rotate(-np.pi / 2, 0, np.pi / 4) \
                    .move(z=0.5) \
                    .material(color)
                if show_axes:
                    ui.scene.line([0, 0, 0], [0.5, 0, 0]).material('#ff0000')
                    ui.scene.line([0, 0, 0], [0, 0.5, 0]).material('#00ff00')
                    ui.scene.line([0, 0, 0], [0, 0, 0.5]).material('#0000ff')
                if show_name:
                    ui.scene.text(camera.name)
            if camera.calibration is not None:
                pyramid.scale(
                    camera.calibration.intrinsics.size.width / px_per_m,
                    camera.calibration.intrinsics.size.height / px_per_m,
                    camera.calibration.intrinsics.matrix[0][0] / px_per_m,
                )



import asyncio
import logging
from asyncio import Task
from collections.abc import AsyncGenerator, Awaitable, Callable

import httpx

from ... import rosys
from ...rosys import on_startup
from ..image_processing import remove_exif
from .vendors import mac_to_url


class MjpegDevice:

    def __init__(self, mac: str, ip: str, *,
                 index: int | None = None,
                 username: str | None = None,
                 password: str | None = None,
                 on_new_image_data: Callable[[bytes, float], Awaitable | None]) -> None:
        self._mac = mac
        self._ip = ip
        self.log = logging.getLogger('rosys.vision.mjpeg_camera.mjpeg_device.' + self._mac)

        self._on_new_image_data = on_new_image_data
        self._capture_task: Task | None = None
        self._authentication = None if username is None or password is None else httpx.DigestAuth(username, password)
        url = mac_to_url(mac, ip, index=index)
        if url is None:
            raise ValueError(f'could not determine URL for {mac}')
        self._url = url

        self.start_capture_task()

    @property
    def is_connected(self) -> bool:
        return (self._capture_task is not None) and (not self._capture_task.done())

    def start_capture_task(self) -> None:
        def create_capture_task() -> None:
            loop = asyncio.get_event_loop()
            self._capture_task = loop.create_task(self.run_capture_task())
        on_startup(create_capture_task)

    def restart_capture(self) -> None:
        self.log.debug('Restarting capture task')
        self.shutdown()
        self.start_capture_task()

    async def run_capture_task(self) -> None:
        self.log.debug('Starting capture task for %s', self._url)

        async def stream() -> AsyncGenerator[bytes, None]:
            async with httpx.AsyncClient() as client:
                assert self._url is not None
                try:
                    async with client.stream('GET', self._url, auth=self._authentication) as response:  # type: ignore
                        if response.status_code != 200:
                            self.log.error('could not connect to %s (credentials: %s): %s %s',
                                           self._url, self._authentication, response.status_code, response.reason_phrase)
                            return

                        buffer_size = 16 * 1024 * 1024
                        buffer = bytearray(buffer_size)
                        buffer_view = memoryview(buffer)
                        buffer_end = 0

                        try:
                            async for chunk in response.aiter_bytes():
                                chunk_len = len(chunk)

                                if buffer_end + chunk_len > buffer_size:
                                    self.log.warning('Buffer overflow, resetting buffer')
                                    buffer_end = 0

                                buffer_view[buffer_end:buffer_end + chunk_len] = chunk
                                buffer_end += chunk_len

                                end = buffer.rfind(b'\xff\xd9', 0, buffer_end)
                                if end == -1:
                                    continue

                                start = buffer.rfind(b'\xff\xd8', 0, end)
                                if start == -1:
                                    continue

                                end += 2
                                yield buffer_view[start:end]
                                buffer_view[:buffer_end - end] = buffer_view[end:buffer_end]
                                buffer_end -= end

                            self.log.debug('Stream ended')
                        except httpx.ReadTimeout:
                            self.log.warning('Connection to %s timed out', self._url)
                except Exception as e:
                    self.log.warning('Connection to %s failed. Was something disconnected?\n%s', self._url, e)
                    raise e

        async for image in stream():
            if not image:
                continue
            try:
                timestamp = rosys.time()
                result = self._on_new_image_data(remove_exif(image), timestamp)
                if isinstance(result, Awaitable):
                    await result
            except Exception as e:
                self.log.error('Error processing image: %s', e)

        self.log.warning('Capture task stopped')
        self._capture_task = None

    def shutdown(self) -> None:
        self.log.debug('Shutting down capture task')
        if self._capture_task is not None:
            self._capture_task.cancel()
            self._capture_task = None

    async def get_fps(self) -> int:
        return 0

    async def set_fps(self, fps: int) -> None:
        pass

    async def get_resolution(self) -> tuple[int, int]:
        return 0, 0

    async def set_resolution(self, width: int, height: int) -> None:
        pass

    async def get_mirrored(self) -> bool:
        return False

    async def set_mirrored(self, mirrored: bool) -> None:
        pass



import asyncio
import gc
import logging
import multiprocessing
import os
import signal
import threading
import time as pytime
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Any, ClassVar, Literal

import numpy as np
import psutil
from nicegui import Client, app, background_tasks, ui

from . import event, run
from .config import Config
from .geometry.frame3d_registry import frame_registry
from .helpers import invoke, is_stopping
from .helpers import is_test as is_test_
from .persistence.registry import backup, restore, sync_backup

log = logging.getLogger('rosys.core')

config = Config()
translator: Any | None = None

is_test = is_test_()

# POSIX standard to create processes is "fork", which is inherently broken for python (see https://pythonspeed.com/articles/python-multiprocessing/)
if __name__ == '__main__':
    multiprocessing.set_start_method('spawn')


@dataclass(slots=True, kw_only=True)
class Notification:
    time: float
    message: str


NEW_NOTIFICATION = event.Event()
"""notify the user (string argument: message)"""


class _state:
    start_time: float = 0.0 if is_test else pytime.time()
    time = start_time
    last_time_request: float = start_time
    exception: BaseException | None = None  # NOTE: used for tests
    startup_finished: bool = False


def get_last_exception() -> BaseException | None:
    return _state.exception


notifications: list[Notification] = []
startup_handlers: list[Callable] = []
shutdown_handlers: list[Callable] = []
tasks: list[asyncio.Task] = []


def notify(message: str,
           type: Literal[  # pylint: disable=redefined-builtin
               'positive',
               'negative',
               'warning',
               'info',
               'ongoing',
           ] | None = None) -> None:
    log.info(message)
    notifications.append(Notification(time=time(), message=message))
    NEW_NOTIFICATION.emit(message)
    # NOTE show notifications on all pages
    for client in Client.instances.values():
        if not client.has_socket_connection:
            continue
        with client:
            try:
                ui.notify(message, type=type)
            except Exception:
                log.exception('failed to call notify')


time_lock = threading.Lock()


def time() -> float:
    if is_test:
        return _state.time
    with time_lock:
        now = pytime.time()
        _state.time += (now - _state.last_time_request) * config.simulation_speed
        _state.last_time_request = now
        return _state.time


def set_time(value: float) -> None:
    assert is_test, 'only tests can change the time'
    _state.time = value


def uptime() -> float:
    return time() - _state.start_time


async def sleep(seconds: float) -> None:
    if is_test:
        sleep_end_time = time() + seconds
        while time() <= sleep_end_time:
            await asyncio.sleep(0)
    else:
        if config.simulation_speed <= 0:
            config.simulation_speed = 0.01
        scaled_seconds = seconds / config.simulation_speed
        count = int(np.ceil(scaled_seconds))
        if count > 0:
            for _ in range(count):
                await asyncio.sleep(scaled_seconds / count)
        else:
            await asyncio.sleep(0)


def _run_handler(handler: Callable) -> None:
    try:
        result = handler()
        if isinstance(result, Awaitable):
            tasks.append(background_tasks.create(result, name=handler.__qualname__))
    except Exception:
        log.exception('error while starting handler "%s"', handler.__qualname__)


class Repeater:
    tasks: ClassVar[set[asyncio.Task]] = set()

    def __init__(self, handler: Callable, interval: float) -> None:
        self.handler = handler
        self.interval = interval
        self._task: asyncio.Task | None = None

    def start(self) -> None:
        if self.running:
            return
        if _state.startup_finished:
            self._task = background_tasks.create(self._repeat())
            self.tasks.add(self._task)
        elif self.start not in startup_handlers:
            startup_handlers.append(self.start)

    async def _repeat(self) -> None:
        await sleep(self.interval)  # NOTE delaying first execution so not all actors rush in at the same time
        while True:
            start = time()
            try:
                if is_stopping():
                    log.info('%s must be stopped', self.handler)
                    break
                await invoke(self.handler)
                dt = time() - start
            except (asyncio.CancelledError, GeneratorExit):
                return
            except Exception:
                dt = time() - start
                log.exception('error in "%s"', self.handler.__qualname__)
                if self.interval == 0 and dt < 0.1:
                    delay = 0.1 - dt
                    log.warning(
                        f'"{self.handler.__qualname__}" would be called to frequently ' +
                        f'because it only took {dt*1000:.0f} ms; ' +
                        f'delaying this step for {delay*1000:.0f} ms')
                    await sleep(delay)
            try:
                await sleep(self.interval - dt)
            except (asyncio.CancelledError, GeneratorExit):
                return

    def stop(self) -> None:
        if not self._task:
            return

        if not self._task.done():
            self._task.cancel()

        self.tasks.remove(self._task)
        self._task = None

    @property
    def running(self) -> bool:
        return self._task is not None and not self._task.done()

    @staticmethod
    def stop_all() -> None:
        for repeater in Repeater.tasks:
            repeater.cancel()


def on_repeat(handler: Callable, interval: float) -> Repeater:
    repeater = Repeater(handler, interval)
    repeater.start()
    return repeater


def on_startup(handler: Callable) -> None:
    if _state.startup_finished:
        _run_handler(handler)
    else:
        startup_handlers.append(handler)


def on_shutdown(handler: Callable) -> None:
    shutdown_handlers.append(handler)


async def startup() -> None:
    if _state.startup_finished:
        raise RuntimeError('should be only executed once')
    if multiprocessing.get_start_method() != 'spawn':
        raise RuntimeError(
            'multiprocessing start method must be "spawn"; see https://pythonspeed.com/articles/python-multiprocessing/')

    restore()

    _state.startup_finished = True

    for handler in startup_handlers:
        _run_handler(handler)

    for coroutine in event.startup_coroutines:
        await coroutine


async def _garbage_collection(mbyte_limit: float = 300) -> None:
    if psutil.virtual_memory().free < mbyte_limit * 1_000_000:
        log.warning('less than %s mb of memory remaining -> start garbage collection', mbyte_limit)
        gc.collect()
        await sleep(1)  # NOTE ensure all warnings appear before sending "finished" message
        log.warning('finished garbage collection')


async def _watch_emitted_events() -> None:
    try:
        for task in event.tasks:
            if task.done() and task.exception():
                _state.exception = task.exception()
                log.exception('task failed to execute', exc_info=task.exception())
        event.tasks = [t for t in event.tasks if not t.done()]
    except Exception:
        log.exception('failed to watch emitted events')


async def shutdown() -> None:
    for handler in shutdown_handlers:
        log.debug('invoking shutdown handler "%s"', handler.__qualname__)
        await invoke(handler)
    log.debug('creating data backup')
    sync_backup(force=True)
    log.debug('tear down "run" tasks')
    run.tear_down()
    log.debug('stopping all repeaters')
    Repeater.stop_all()
    log.debug('canceling all remaining tasks')
    for task in tasks:
        task.cancel()
    log.debug('clearing tasks')
    tasks.clear()

    # NOTE: kill own process and all its children after uvicorn has finished (we had many restart freezes before)
    def delayed_kill():
        pytime.sleep(1)
        os.kill(os.getpid(), signal.SIGKILL)
    if not is_test:
        threading.Thread(target=delayed_kill).start()
    log.debug('finished shutdown')


def reset_before_test() -> None:
    assert is_test
    set_time(0)  # NOTE: in tests we start at zero for better readability
    _state.exception = None


def reset_after_test() -> None:
    assert is_test
    startup_handlers.clear()
    tasks.clear()
    _state.startup_finished = False
    shutdown_handlers.clear()
    event.reset()
    frame_registry.clear()

    register_base_startup_handlers()


def register_base_startup_handlers() -> None:
    on_repeat(_garbage_collection, 60)
    on_repeat(_watch_emitted_events, 0.1)
    on_repeat(backup, 10)


gc.disable()  # NOTE disable automatic garbage collection to optimize performance
register_base_startup_handlers()

app.on_startup(startup)
app.on_shutdown(shutdown)



import logging
from typing import Any

import numpy as np
from typing_extensions import Self

from ... import rosys
from ..camera.configurable_camera import ConfigurableCamera
from ..camera.transformable_camera import TransformableCamera
from ..image import Image
from ..image_processing import get_image_size_from_bytes, process_jpeg_image, process_ndarray_image
from ..image_rotation import ImageRotation
from .usb_device import UsbDevice


class UsbCamera(ConfigurableCamera, TransformableCamera):

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: str | None = None,
                 connect_after_init: bool = True,
                 auto_exposure: bool = True,
                 exposure: bool = False,
                 width: int = 800,
                 height: int = 600,
                 fps: int = 10,
                 **kwargs) -> None:
        super().__init__(id=id,
                         name=name,
                         connect_after_init=connect_after_init,
                         **kwargs)
        self._pending_operations = 0
        self.device: UsbDevice | None = None
        self.detect: bool = False
        self.color: str | None = None

        self._register_parameter('auto_exposure', self.get_exposure, self.set_exposure, auto_exposure)
        self._register_parameter('exposure', self.get_exposure, self.set_exposure, exposure)
        self._register_parameter('width', self.get_width, self.set_width, width)
        self._register_parameter('height', self.get_height, self.set_height, height)
        self._register_parameter('fps', self.get_fps, self.set_fps, fps)

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {
            name: param.value for name, param in self._parameters.items()
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)

    @property
    def is_connected(self) -> bool:
        return self.device is not None

    async def connect(self) -> None:
        if self.is_connected:
            return

        device = UsbDevice.from_uid(self.id, self._handle_new_image_data)
        if device is None:
            logging.warning('Connecting camera %s: failed', self.id)
            return

        self.device = device
        logging.info('Connecting camera %s: succeeded', self.id)

        await self._apply_all_parameters()

    async def disconnect(self) -> None:
        if not self.is_connected:
            return

        assert self.device is not None
        await self.device.release_capture()
        self.device = None
        logging.info('camera %s: disconnected', self.id)

    async def _handle_new_image_data(self, image_data: np.ndarray | bytes, timestamp: float) -> None:
        if not self.is_connected:
            return None

        assert self.device is not None

        bytes_: bytes | None
        if isinstance(image_data, np.ndarray):
            bytes_ = await rosys.run.cpu_bound(process_ndarray_image, image_data, self.rotation, self.crop)
        else:
            bytes_ = image_data
            if self.crop or self.rotation != ImageRotation.NONE:
                bytes_ = await rosys.run.cpu_bound(process_jpeg_image, bytes_, self.rotation, self.crop)
        if bytes_ is None:
            return

        final_image_resolution = get_image_size_from_bytes(bytes_)

        image = Image(time=timestamp, camera_id=self.id, size=final_image_resolution, data=bytes_)
        self._add_image(image)

    def set_auto_exposure(self, auto: bool) -> None:
        assert self.device is not None
        self.device.set_auto_exposure(auto)
        if not auto:
            manual_exposure = self._parameters['exposure'].value
            assert manual_exposure is not None
            self.device.set_exposure(manual_exposure)

    def set_exposure(self, value: float) -> None:
        assert self.device is not None
        self.device.set_exposure(value)

    def get_auto_exposure(self) -> bool | None:
        assert self.device is not None
        return self.device.get_auto_exposure()

    def get_exposure(self) -> float | None:
        assert self.device is not None
        return self.device.get_exposure()

    def set_width(self, width: int) -> None:
        assert self.device is not None
        self.device.set_width(width)

    def get_width(self) -> int:
        assert self.device is not None
        return self.device.get_width()

    def set_height(self, height: int) -> None:
        assert self.device is not None
        self.device.set_height(height)

    def get_height(self) -> int:
        assert self.device is not None
        return self.device.get_height()

    def set_fps(self, fps: int) -> None:
        assert self.device is not None
        self.device.set_fps(fps)

    def get_fps(self) -> int:
        assert self.device is not None
        return self.device.get_fps()



import asyncio
import logging
import shlex
import signal
import subprocess
from asyncio.subprocess import Process
from collections.abc import AsyncGenerator, Awaitable, Callable
from io import BytesIO

from nicegui import background_tasks

from ... import rosys
from .jovision_rtsp_interface import JovisionInterface
from .vendors import VendorType, mac_to_url, mac_to_vendor


class RtspDevice:

    def __init__(self, mac: str, ip: str, *,
                 jovision_profile: int, fps: int, on_new_image_data: Callable[[bytes, float], Awaitable | None]) -> None:
        self._mac = mac
        self._ip = ip
        self.log = logging.getLogger('rosys.vision.rtsp_camera.rtsp_device.' + self._mac)

        self._fps = fps
        self._jovision_profile = jovision_profile
        self._on_new_image_data = on_new_image_data

        self._capture_task: asyncio.Task | None = None
        self._capture_process: Process | None = None
        self._authorized: bool = True

        vendor_type = mac_to_vendor(mac)

        self._settings_interface: JovisionInterface | None = None
        if vendor_type == VendorType.JOVISION:
            self._settings_interface = JovisionInterface(ip)
        else:
            self.log.warning('[%s] No settings interface for vendor type %s', self._mac, vendor_type)
            self.log.warning('[%s] Using default fps of 10', self._mac)

        self.log.info('[%s] Starting VideoStream for %s', self._mac, self.url)
        self._start_gstreamer_task()

    @property
    def is_connected(self) -> bool:
        return self._capture_task is not None

    @property
    def authorized(self) -> bool:
        return self._authorized

    @property
    def url(self) -> str:
        url = mac_to_url(self._mac, self._ip, self._jovision_profile)
        if url is None:
            raise ValueError(f'could not determine RTSP URL for {self._mac}')
        return url

    async def shutdown(self) -> None:
        if self._capture_process is not None:
            self.log.debug('[%s] Terminating gstreamer process', self._mac)
            self._capture_process.terminate()
            try:
                await asyncio.wait_for(self._capture_process.wait(), timeout=5)
            except asyncio.TimeoutError:
                self.log.warning('[%s] Timeout while waiting for gstreamer process to terminate', self._mac)
            else:
                self.log.debug('[%s] Successfully shut down process (code %s)',
                               self._mac, self._capture_process.returncode if self._capture_process.returncode is not None else 'None')
                self._capture_process = None
        if self._capture_task is not None and not self._capture_task.done():
            self.log.debug('[%s] Cancelling gstreamer task', self._mac)
            self._capture_task.cancel()
            try:
                await asyncio.wait_for(self._capture_task, timeout=5)
            except asyncio.TimeoutError:
                self.log.warning('[%s] Timeout while waiting for capture task to cancel', self._mac)
                return
            except asyncio.CancelledError:
                self.log.debug('[%s] Task was successfully cancelled', self._mac)
            else:
                self.log.debug('[%s] Task finished', self._mac)
            self._capture_task = None

    def _start_gstreamer_task(self) -> None:
        self.log.debug('[%s] Starting gstreamer task', self._mac)
        if self._capture_task is not None and not self._capture_task.done():
            self.log.warning('[%s] capture task already running', self._mac)
            return
        self._capture_task = background_tasks.create(self._run_gstreamer(), name=f'capture {self._mac}')

    async def restart_gstreamer(self) -> None:
        await self.shutdown()
        self._start_gstreamer_task()

    async def _run_gstreamer(self) -> None:
        if self._capture_process is not None and self._capture_process.returncode is None:
            self.log.warning('[%s] capture process already running', self._mac)
            return

        async def stream() -> AsyncGenerator[bytes, None]:
            url = self.url
            if 'subtype=0' in url:
                url = url.replace('subtype=0', 'subtype=1')

            self.log.debug('[%s] Starting gstreamer pipeline for %s', self._mac, url)
            # to try: replace avdec_h264 with nvh264dec ! nvvidconv (!videoconvert)
            command = f'gst-launch-1.0 rtspsrc location="{url}" latency=0 protocols=tcp ! rtph264depay ! avdec_h264 ! videoconvert ! jpegenc ! fdsink'
            self.log.debug('[%s] Running command: %s', self._mac, command)
            process = await asyncio.create_subprocess_exec(
                *shlex.split(command),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                limit=1024*1024*1024,
            )
            assert process.stdout is not None
            assert process.stderr is not None
            self._capture_process = process

            buffer = BytesIO()
            chunk_size = 1024*1024

            while process.returncode is None:
                assert process.stdout is not None

                eof = False
                while True:
                    new = await process.stdout.read(chunk_size)
                    eof = not new
                    buffer.write(new)
                    if len(new) < chunk_size:
                        break

                if eof:
                    break

                data = buffer.getvalue()
                end = data.rfind(b'\xff\xd9')
                if end == -1:
                    continue

                start = data.rfind(b'\xff\xd8', 0, end)
                if start == -1:
                    continue

                end += 2

                yield data[start:end]
                rest = data[end:]
                buffer.seek(0)
                buffer.truncate()
                buffer.write(rest)

            try:
                await asyncio.wait_for(process.wait(), timeout=5)
            except asyncio.TimeoutError:
                self.log.warning(
                    '[%s] Stream ended. Timeout while waiting for gstreamer process to terminate', self._mac)
                return

            return_code = process.returncode
            if return_code == -1 * signal.SIGTERM:
                self.log.debug('gstreamer process %s was terminated using SIGTERM', process.pid)
            else:
                error = await process.stderr.read()
                error_message = error.decode()
                self.log.error('gstreamer process %s exited with code %s.\nstderr: %s',
                               process.pid, return_code, error_message)

                if 'Unauthorized' in error_message:
                    self._authorized = False

        async for image in stream():
            timestamp = rosys.time()
            result = self._on_new_image_data(image, timestamp)
            if isinstance(result, Awaitable):
                await result

        self.log.info('[%s] stream ended', self._mac)

        self._capture_task = None

    async def set_fps(self, fps: int) -> None:
        self._fps = fps

        if self._settings_interface is not None:
            await self._settings_interface.set_fps(stream_id=self._jovision_profile, fps=self._fps)

    async def get_fps(self) -> int | None:
        if self._settings_interface is not None:
            return await self._settings_interface.get_fps(stream_id=self._jovision_profile)
        return self._fps

    def set_jovision_profile(self, profile: int) -> None:
        self._jovision_profile = profile

    def get_jovision_profile(self) -> int:
        return self._jovision_profile

    async def set_bitrate(self, bitrate: int) -> None:
        if self._settings_interface is not None:
            await self._settings_interface.set_bitrate(stream_id=self._jovision_profile, bitrate=bitrate)

    async def get_bitrate(self) -> int | None:
        if self._settings_interface is not None:
            return await self._settings_interface.get_bitrate(stream_id=self._jovision_profile)
        return None



from dataclasses import dataclass, field
from typing import Protocol

import numpy as np

from .. import rosys
from ..analysis import track
from ..geometry import Point, Pose, Spline
from ..helpers import ModificationContext, eliminate_2pi, eliminate_pi, ramp
from .drivable import Drivable
from .odometer import Odometer
from .path_segment import PathSegment


@dataclass(slots=True, kw_only=True)
class DriveParameters(ModificationContext):
    linear_speed_limit: float = 0.5
    angular_speed_limit: float = 0.5
    minimum_turning_radius: float = 0.0
    can_drive_backwards: bool = True
    max_detection_age_ramp: tuple[float, float] | None = None
    hook_offset: float = 0.5
    carrot_offset: float = 0.6
    carrot_distance: float = 0.1
    hook_bending_factor: float = 0


@dataclass(slots=True, kw_only=True)
class DriveState:
    carrot_pose: Pose
    curvature: float
    backward: bool
    turn_angle: float


class DrivingAbortedException(Exception):
    pass


class PoseProvider(Protocol):

    @property
    def pose(self) -> Pose:
        ...


class Driver:
    """The driver module allows following a given path.

    It requires a wheels module (or any drivable hardware representation) to execute individual drive commands.
    It also requires an odometer to get a current prediction of the robot's pose.
    Its `parameters` allow controlling the specific drive behavior.
    """

    def __init__(self, wheels: Drivable, odometer: Odometer | PoseProvider) -> None:
        self.wheels = wheels
        self.odometer = odometer
        self.parameters = DriveParameters()
        self.state: DriveState | None = None
        self._abort = False

    @property
    def prediction(self) -> Pose:
        """The current prediction of the robot's pose based on the odometer."""
        return self.odometer.prediction if isinstance(self.odometer, Odometer) else self.odometer.pose

    def abort(self) -> None:
        """Abort the current drive routine."""
        self._abort = True

    @track
    async def drive_path(self,
                         path: list[PathSegment], *,
                         throttle_at_end: bool = True,
                         stop_at_end: bool = True) -> None:
        """Drive along a given path.

        :param path: The path to drive along, composed of PathSegments.
        :param throttle_at_end: Whether to throttle down when approaching the end of the path (default: ``True``).
        :param stop_at_end: Whether to stop at the end of the path (default: ``True``).
        :raises: DrivingAbortedException: If the driving process is aborted.
        """
        for segment in path:
            await self.drive_spline(segment.spline,
                                    flip_hook=segment.backward,
                                    throttle_at_end=throttle_at_end and segment == path[-1],
                                    stop_at_end=stop_at_end and segment == path[-1])

    @track
    async def drive_to(self,
                       target: Point, *,
                       backward: bool = False,
                       throttle_at_end: bool = True,
                       stop_at_end: bool = True) -> None:
        """Drive to a given target point.

        :param target: The target point to drive to.
        :param backward: Whether to drive backwards (default: ``False``).
        :param throttle_at_end: Whether to throttle down when approaching the target point (default: ``True``).
        :param stop_at_end: Whether to stop at the target point (default: ``True``).
        :raises: DrivingAbortedException: If the driving process is aborted.
        """
        if self.parameters.minimum_turning_radius:
            await self.drive_circle(target, backward=backward, stop_at_end=False)

        robot_position = self.prediction.point
        approach_spline = Spline(
            start=robot_position,
            control1=robot_position.interpolate(target, 1/3),
            control2=robot_position.interpolate(target, 2/3),
            end=target,
        )
        await self.drive_spline(approach_spline, flip_hook=backward, throttle_at_end=throttle_at_end, stop_at_end=stop_at_end)

    @track
    async def drive_circle(self,
                           target: Point, *,
                           angle_threshold: float = np.deg2rad(5),
                           backward: bool = False,
                           stop_at_end: bool = True) -> None:
        """Drive in a circular path.

        When the angle between the robot's current direction and the target direction is less than ``angle_threshold``,
        the robot stops driving.

        :param target: The target point to drive towards.
        :param angle_threshold: The angle threshold to stop driving (radians, default: 5°).
        :param backward: Whether to drive backwards (default: ``False``).
        :param stop_at_end: Whether to stop the robot at the end of the circular path (default: ``False``).
        :raises: DrivingAbortedException: If the driving process is aborted.
        """
        while True:
            if self._abort:
                self._abort = False
                raise DrivingAbortedException()
            target_yaw = self.prediction.direction(target)
            if backward:
                target_yaw += np.pi
            angle = eliminate_2pi(target_yaw - self.prediction.yaw)
            if abs(angle) < angle_threshold:
                break
            linear = 0.5 if not backward else -0.5
            sign = 1 if angle > 0 else -1
            if backward:
                sign *= -1
            angular = linear / self.parameters.minimum_turning_radius * sign
            await self.wheels.drive(*self._throttle(linear, angular))
            await rosys.sleep(0.1)
        if stop_at_end:
            await self.wheels.stop()

    @track
    async def drive_spline(self,
                           spline: Spline, *,
                           flip_hook: bool = False,
                           throttle_at_end: bool = True,
                           stop_at_end: bool = True) -> None:
        """Drive along a given spline.

        :param spline: The spline to drive along.
        :param flip_hook: Whether to flip the hook offset (default: ``False``).
        :param throttle_at_end: Whether to throttle down when approaching the end of the spline (default: ``True``).
        :param stop_at_end: Whether to stop at the end of the spline (default: ``True``).
        :raises DrivingAbortedException: If the driving process is aborted.
        """
        if spline.start.distance(spline.end) < 0.01:
            return  # NOTE: skip tiny splines

        hook_offset = Point(x=self.parameters.hook_offset, y=0) * (-1 if flip_hook else 1)
        carrot_offset = Point(x=self.parameters.carrot_offset, y=0)
        carrot = Carrot(spline=spline, offset=carrot_offset)

        while True:
            if self._abort:
                self._abort = False
                raise DrivingAbortedException()
            velocity = self.odometer.current_velocity if isinstance(self.odometer, Odometer) else None
            dYaw = self.parameters.hook_bending_factor * velocity.angular if velocity else 0
            hook = self.prediction.transform_pose(Pose(yaw=dYaw)).transform(hook_offset)
            if self.parameters.can_drive_backwards:
                can_move = carrot.move(hook, distance=self.parameters.carrot_distance)
            else:
                can_move = carrot.move_by_foot(self.prediction)
            if not can_move:
                break

            turn_angle = eliminate_pi(hook.direction(carrot.offset_point) - self.prediction.yaw)
            curvature = np.tan(turn_angle) / hook_offset.x
            if curvature != 0 and abs(1 / curvature) < self.parameters.minimum_turning_radius:
                curvature = (-1 if curvature < 0 else 1) / self.parameters.minimum_turning_radius

            drive_backward = hook.projected_distance(carrot.offset_point, self.prediction.yaw) < 0
            if drive_backward and not self.parameters.can_drive_backwards:
                drive_backward = False
                curvature = (-1 if curvature > 0 else 1) / max(self.parameters.minimum_turning_radius, 0.001)
            linear: float = -1 if drive_backward else 1
            t = spline.closest_point(hook.x, hook.y)
            if t >= 1.0 and throttle_at_end:
                target_distance = self.prediction.projected_distance(spline.pose(1.0))
                linear *= ramp(target_distance, self.parameters.hook_offset, 0.0, 1.0, 0.01, clip=True)
            angular = linear * curvature

            self.state = DriveState(
                carrot_pose=carrot.pose,
                curvature=curvature,
                backward=drive_backward,
                turn_angle=turn_angle,
            )

            await self.wheels.drive(*self._throttle(linear, angular))
            await rosys.sleep(0.1)

        self.state = None
        if stop_at_end:
            await self.wheels.stop()

    def _throttle(self, linear: float, angular: float) -> tuple[float, float]:
        factor = self.throttle_factor()
        linear *= factor
        angular *= factor

        linear_limit = self.parameters.linear_speed_limit
        angular_limit = self.parameters.angular_speed_limit

        if abs(linear) > linear_limit:
            factor = linear_limit / abs(linear)
            angular *= factor
            linear *= factor
        if abs(angular) > angular_limit:
            factor = angular_limit / abs(angular)
            linear *= factor
            angular *= factor

        return linear, angular

    def throttle_factor(self) -> float:
        if self.parameters.max_detection_age_ramp is None:
            return 1
        if not isinstance(self.odometer, Odometer):
            raise ValueError('max_detection_age_ramp requires an odometer')
        if self.odometer.detection is None:
            return 0
        age_ramp = self.parameters.max_detection_age_ramp
        age = rosys.time() - self.odometer.detection.time
        return ramp(age, age_ramp[0], age_ramp[1], 1.0, 0.0, clip=True)


@dataclass(slots=True, kw_only=True)
class Carrot:
    spline: Spline
    offset: Point = field(default_factory=lambda: Point(x=0, y=0))
    t: float = 0
    _estimated_spline_length: float = field(init=False)

    def __post_init__(self) -> None:
        self._estimated_spline_length = self.spline.estimated_length()  # NOTE: compute once to avoid repeated computation

    @property
    def pose(self) -> Pose:
        if self.t < 1.0:
            return self.spline.pose(self.t)
        return Pose(
            x=self.spline.x(1.0) + (self.t - 1.0) * self.spline.gx(1.0),
            y=self.spline.y(1.0) + (self.t - 1.0) * self.spline.gy(1.0),
            yaw=self.spline.yaw(1.0),
        )

    @property
    def offset_point(self) -> Point:
        return self.pose.transform(self.offset)

    def move(self, hook: Point, distance: float) -> bool:
        dt = 0.1 * distance / self._estimated_spline_length
        while hook.distance(self.offset_point) < distance:
            self.t += dt
            if self.t >= 1.0:
                return False
        return True

    def move_by_foot(self, pose: Pose) -> bool:
        self.t = self.spline.closest_point(pose.x, pose.y, t_min=self.t, t_max=1.0)
        return self.t < 1.0



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import ClassVar

from .point import Point
from .pose import Pose

RADIUS = 6371000.0
ZERO_POINT = Point(x=0, y=0)


@dataclass(slots=True)
class GeoPoint:
    lat: float
    lon: float

    @classmethod
    def from_degrees(cls, lat: float, lon: float) -> GeoPoint:
        """Create a geo point from latitude and longitude in degrees."""
        return cls(math.radians(lat), math.radians(lon))

    @classmethod
    def from_point(cls, point: Point) -> GeoPoint:
        """Create a geo point from a local point in the current geo reference."""
        assert GeoReference.current is not None
        return GeoReference.current.point_to_geo(point)

    def direction(self, other: GeoPoint | GeoPose) -> float:
        """Calculate the direction to another geo point."""
        return math.atan2(math.sin(other.lon - self.lon) * math.cos(other.lat),
                          math.cos(self.lat) * math.sin(other.lat) -
                          math.sin(self.lat) * math.cos(other.lat) * math.cos(other.lon - self.lon))

    def distance(self, other: GeoPoint | GeoPose) -> float:
        """Calculate the distance to another geo point in meters using the Haversine formula."""
        d_lat = other.lat - self.lat
        d_lon = other.lon - self.lon
        a = math.sin(d_lat / 2)**2 + math.cos(self.lat) * math.cos(other.lat) * math.sin(d_lon / 2)**2
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
        return RADIUS * c

    def polar(self, distance: float, direction: float) -> GeoPoint:
        """Calculate the geo point at a given distance and direction from this geo point."""
        angular_distance = distance / RADIUS
        lat2 = math.asin(math.sin(self.lat) * math.cos(angular_distance) +
                         math.cos(self.lat) * math.sin(angular_distance) * math.cos(direction))
        lon2 = self.lon + math.atan2(math.sin(direction) * math.sin(angular_distance) * math.cos(self.lat),
                                     math.cos(angular_distance) - math.sin(self.lat) * math.sin(lat2))
        return GeoPoint(lat2, lon2)

    def shift_by(self, *, x: float = 0.0, y: float = 0.0) -> GeoPoint:
        """Shift the geo point by Cartesian coordinates in meters relative to the current geo reference.

        If no current geo reference is set, x will be applied in North direction and y in West direction.
        """
        distance = math.sqrt(x**2 + y**2)
        angle = math.atan2(-y, x) + GeoReference.current.direction if GeoReference.current is not None else 0
        return self.polar(distance, angle)

    def to_local(self) -> Point:
        """Transform to local Cartesian coordinates relative to the current geo reference."""
        assert GeoReference.current is not None
        return GeoReference.current.point_to_local(self)

    def __str__(self) -> str:
        lat_deg, lon_deg = self.degree_tuple
        return f'GeoPoint(lat={lat_deg:.6f}˚, lon={lon_deg:.6f}˚)'

    @property
    def degree_tuple(self) -> tuple[float, float]:
        """Latitude and longitude in degrees."""
        return math.degrees(self.lat), math.degrees(self.lon)

    @property
    def tuple(self) -> tuple[float, float]:
        """Latitude and longitude in radians."""
        return self.lat, self.lon


@dataclass(slots=True)
class GeoPose:
    lat: float
    lon: float
    heading: float

    @classmethod
    def from_degrees(cls, lat: float, lon: float, heading: float) -> GeoPose:
        """Create a geo pose from latitude, longitude and heading in degrees."""
        return cls(math.radians(lat), math.radians(lon), math.radians(heading))

    @classmethod
    def from_pose(cls, pose: Pose) -> GeoPose:
        """Create a geo pose from a pose in the current geo reference."""
        assert GeoReference.current is not None
        geo_point1 = GeoReference.current.point_to_geo(pose)
        geo_point2 = GeoReference.current.point_to_geo(pose.transform_pose(Pose(x=1)))
        return cls(geo_point1.lat, geo_point1.lon, geo_point1.direction(geo_point2))

    @property
    def point(self) -> GeoPoint:
        """The geo point of the geo pose."""
        return GeoPoint(self.lat, self.lon)

    def to_local(self) -> Pose:
        """Transform to local Cartesian coordinates relative to the current geo reference."""
        assert GeoReference.current is not None
        point1 = GeoReference.current.point_to_local(self)
        point2 = GeoReference.current.point_to_local(self.point.polar(1, self.heading))
        return Pose(x=point1.x, y=point1.y, yaw=point1.direction(point2))

    def __str__(self) -> str:
        lat_deg, lon_deg, heading_deg = self.degree_tuple
        return f'GeoPose(lat={lat_deg:.6f}˚, lon={lon_deg:.6f}˚, heading={heading_deg:.1f}˚)'

    @property
    def degree_tuple(self) -> tuple[float, float, float]:
        """Latitude, longitude, and heading in degrees."""
        return math.degrees(self.lat), math.degrees(self.lon), math.degrees(self.heading)

    @property
    def tuple(self) -> tuple[float, float, float]:
        """Latitude, longitude, and heading in radians."""
        return self.lat, self.lon, self.heading


@dataclass(slots=True)
class Fixpoint:
    local_point: Point
    geo_point: GeoPoint | None = None


@dataclass(slots=True)
class GeoReference:
    current: ClassVar[GeoReference | None] = None
    origin: GeoPoint
    direction: float = 0.0  # direction of the local x-axis in the global geo system (0 = north, pi/2 = east)

    @classmethod
    def update_current(cls, new_reference: GeoReference) -> None:
        """Update the current geo reference to the given reference."""
        if cls.current is None:
            cls.current = new_reference
        else:
            cls.current.origin = new_reference.origin
            cls.current.direction = new_reference.direction

    @classmethod
    def from_two_fixpoints(cls, A: Fixpoint, B: Fixpoint) -> GeoReference:
        """Create a geo reference from two fixpoints.

        :param A: The first fixpoint.
        :param B: The second fixpoint.
        :raises: AssertionError: If either fixpoint does not have a geo_point.
        """
        assert A.geo_point is not None, 'Fixpoint "A" must have a geo_point'
        assert B.geo_point is not None, 'Fixpoint "B" must have a geo_point'
        x_direction = A.geo_point.direction(B.geo_point) + A.local_point.direction(B.local_point)
        origin = A.geo_point \
            .polar(A.local_point.x, x_direction + math.radians(180)) \
            .polar(A.local_point.y, x_direction + math.radians(90))
        return cls(origin, x_direction)

    def point_to_geo(self, point: Point | Pose) -> GeoPoint:
        """Convert a local point to a global point."""
        return self.origin.polar(ZERO_POINT.distance(point), self.direction - ZERO_POINT.direction(point))

    def pose_to_geo(self, pose: Pose) -> GeoPose:
        """Convert a local pose to a global geo pose."""
        geo_point1 = self.point_to_geo(pose)
        geo_point2 = self.point_to_geo(pose.transform_pose(Pose(x=1)))
        return GeoPose(geo_point1.lat, geo_point1.lon, geo_point1.direction(geo_point2))

    def point_to_local(self, geo_point: GeoPoint | GeoPose) -> Point:
        """Convert a global point to a local point."""
        return ZERO_POINT.polar(self.origin.distance(geo_point), self.direction - self.origin.direction(geo_point))

    def pose_to_local(self, geo_pose: GeoPose) -> Pose:
        """Convert a global geo pose to a local pose."""
        point1 = self.point_to_local(geo_pose)
        point2 = self.point_to_local(geo_pose.point.polar(1, geo_pose.heading))
        return Pose(x=point1.x, y=point1.y, yaw=point1.direction(point2))

    @property
    def degree_tuple(self) -> tuple[float, float, float]:
        """Latitude, longitude, and direction of the x-axis in degrees."""
        return (*self.origin.degree_tuple, math.degrees(self.direction))

    @property
    def tuple(self) -> tuple[float, float, float]:
        """Latitude, longitude, and direction of the x-axis in radians."""
        return (*self.origin.tuple, self.direction)

    def __str__(self) -> str:
        lat_deg, lon_deg, direction_deg = self.degree_tuple
        return f'GeoReference(lat={lat_deg:.6f}˚, lon={lon_deg:.6f}˚, heading={direction_deg:.1f}˚)'



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field

import numpy as np
from typing_extensions import Self

from .frame3d_registry import frame_registry
from .object3d import Object3d
from .point3d import Point3d
from .rotation import Rotation


@dataclass(slots=True, kw_only=True)
class Pose3d(Object3d):
    """A 3D pose consisting of a translation and a rotation.

    The pose is stored as the transformation to the parent frame (not the objects in it).
    The transformation first applies the rotation and then the translation.
    """
    x: float = 0
    y: float = 0
    z: float = 0
    rotation: Rotation = field(default_factory=Rotation.zero)

    @classmethod
    def zero(cls) -> Pose3d:
        return Pose3d(x=0, y=0, z=0, rotation=Rotation.zero())

    def as_frame(self, frame_id: str) -> Frame3d:
        """Register this pose as a frame."""
        return Frame3d(_frame_id=self._frame_id,
                       x=self.x,
                       y=self.y,
                       z=self.z,
                       rotation=self.rotation,
                       id=frame_id)

    @classmethod
    def from_matrix(cls, M: np.ndarray) -> Pose3d:
        return cls(x=M[0, 3] / M[3, 3], y=M[1, 3] / M[3, 3], z=M[2, 3] / M[3, 3],
                   rotation=Rotation(R=(M[:3, :3] / M[3, 3]).tolist()))

    @property
    def point_3d(self) -> Point3d:
        return Point3d(x=self.x, y=self.y, z=self.z)

    @property
    def translation(self) -> tuple[float, float, float]:
        return self.x, self.y, self.z

    @property
    def translation_vector(self) -> np.ndarray:
        return np.array(self.translation).reshape(3, 1)

    @property
    def matrix(self) -> np.ndarray:
        return np.block([[self.rotation.matrix, self.translation_vector], [0, 0, 0, 1]])

    @property
    def inverse_matrix(self) -> np.ndarray:
        return np.block([[self.rotation.T.matrix, -self.rotation.matrix.T @ self.translation_vector], [0, 0, 0, 1]])

    def inverse(self) -> Pose3d:
        return Pose3d.from_matrix(self.inverse_matrix)

    def __matmul__(self, other: Pose3d) -> Pose3d:
        return Pose3d.from_matrix(self.matrix @ other.matrix)

    def __str__(self) -> str:
        return f'T=({self.x}, {self.y}, {self.z})\nR = {self.rotation}'

    def transform_with(self, pose: Pose3d) -> Pose3d:
        """Transform this pose with another pose."""
        return pose @ self


@dataclass(slots=True, kw_only=True)
class Frame3d(Pose3d):
    id: str | None = None

    def __post_init__(self) -> None:
        if self.id is not None:
            frame_registry[self.id] = self

    def in_frame(self, frame: Frame3d | None) -> Self:
        if frame is not None and self in frame.ancestors:
            raise ValueError(f'Cannot place frame "{self.id}" in frame "{frame.id}" because it creates a cycle')
        self._frame_id = frame.id if frame is not None else None
        return self

    @property
    def ancestors(self) -> Iterator[Frame3d]:
        yield self
        if self.frame_id:
            yield from frame_registry[self.frame_id].ancestors



from __future__ import annotations

import logging
import math
from abc import ABC
from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

import numpy as np
import serial
from nicegui import ui
from serial.tools import list_ports

from .. import rosys
from ..event import Event
from ..geometry import GeoPoint, GeoPose, Pose
from ..run import io_bound

if TYPE_CHECKING:
    from ..hardware import WheelsSimulation


class GpsQuality(IntEnum):
    INVALID = 0
    GPS = 1
    DGPS = 2
    PPS = 3
    RTK_FIXED = 4
    RTK_FLOAT = 5
    DEAD_RECKONING = 6
    MANUAL = 7
    SIMULATION = 8


@dataclass
class GnssMeasurement:
    time: float
    pose: GeoPose
    latitude_std_dev: float = 0.0
    longitude_std_dev: float = 0.0
    heading_std_dev: float = 0.0
    gps_quality: GpsQuality = GpsQuality.INVALID
    num_satellites: int = 0
    hdop: float = 0.0
    altitude: float = 0.0

    @property
    def point(self) -> GeoPoint:
        return self.pose.point

    @property
    def heading(self) -> float:
        return self.pose.heading


class Gnss(ABC):
    """A GNSS module that provides measurements from a GNSS receiver."""

    def __init__(self) -> None:
        self.log = logging.getLogger('rosys.gnss')
        self.last_measurement: GnssMeasurement | None = None

        self.NEW_MEASUREMENT = Event()
        """a new measurement has been received (argument: ``GnssMeasurement``)"""

    @property
    def is_connected(self) -> bool:
        return False

    def developer_ui(self) -> None:
        ui.label('GNSS').classes('text-center text-bold')
        with ui.grid(columns='auto auto').classes('gap-y-1 w-2/5'):
            ui.label('Connected:')
            ui.icon('close').bind_name_from(self, 'is_connected',
                                            lambda x: 'check' if x else 'close')
            ui.label('Position:')
            with ui.column().classes('gap-y-0'):
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{math.degrees(x.pose.lat):.8f}˚' if x else '-')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'± {x.latitude_std_dev:.3f}m' if x else '-')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{math.degrees(x.pose.lon):.8f}˚' if x else '-')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'± {x.longitude_std_dev:.3f}m' if x else '-')
            ui.label('Heading:')
            with ui.column().classes('gap-y-0'):
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{math.degrees(x.pose.heading):.2f}˚' if x else '-')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'± {x.heading_std_dev:.2f}˚' if x else '-')
            ui.label('Quality:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: x.gps_quality.name if x else '')
            ui.label('Satellites:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: str(x.num_satellites) if x else '-')
            ui.label('HDOP:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: f'{x.hdop:.2f}' if x else '-')
            ui.label('Altitude:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: f'{x.altitude:.3f}m' if x else '-')
            ui.label('Last update:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: f'{rosys.time() - x.time:.2f}s' if x else '')


class GnssHardware(Gnss):
    """This hardware module connects to a Septentrio SimpleRTK3b (Mosaic-H) GNSS receiver."""

    def __init__(self, *, antenna_pose: Pose | None) -> None:
        """
        :param antenna_pose: the pose of the main antenna in the robot's coordinate frame (yaw: direction to the auxiliary antenna)
        """
        super().__init__()
        self.antenna_pose = antenna_pose or Pose(x=0.0, y=0.0, yaw=0.0)
        self._antenna_distance = math.sqrt(self.antenna_pose.x**2 + self.antenna_pose.y**2)
        """the distance from the robot's center to the main antenna"""
        self._antenna_angle = math.pi + math.atan2(self.antenna_pose.y, self.antenna_pose.x) - self.antenna_pose.yaw
        """the angle from the robot's center to the main antenna"""

        serial_device_path = self._find_device()
        self.serial_connection = self._connect_to_device(serial_device_path)
        rosys.on_startup(self._run)

    @property
    def is_connected(self) -> bool:
        if self.serial_connection is None:
            self.log.debug('Device not connected')
            return False
        if not self.serial_connection.isOpen():
            self.log.debug('Device not open')
            return False
        return True

    async def _run(self) -> None:
        buffer = ''
        last_raw_latitude = 0.0
        last_raw_longitude = 0.0
        last_raw_heading = 0.0
        last_latitude_accuracy = 0.0
        last_longitude_accuracy = 0.0
        last_heading_accuracy = 0.0
        last_gps_quality = 0
        last_num_satellites = 0
        last_hdop = 0.0
        last_altitude = 0.0
        last_gga_timestamp = ''
        last_gst_timestamp = ''
        last_pssn_timestamp = ''
        while True:
            if not self.is_connected:
                return None
            result = await io_bound(self.serial_connection.read_until, b'\r\n')
            if not result:
                self.log.debug('No data')
                continue
            line = result.decode('utf-8')
            if line.endswith('\r\n'):
                sentence = buffer + line.split('*')[0]
                buffer = ''
            else:
                buffer += line
                continue

            self.log.debug(sentence)
            parts = sentence.split(',')
            try:
                timestamp = parts[{'$GPGGA': 1, '$GPGST': 1, '$PSSN': 2}[parts[0]]]
                if parts[0] == '$GPGGA':
                    if parts[2] == '' or parts[4] == '':
                        continue
                    last_gga_timestamp = timestamp
                    last_raw_latitude = self._convert_to_decimal(parts[2], parts[3])
                    last_raw_longitude = self._convert_to_decimal(parts[4], parts[5])
                    last_gps_quality = int(parts[6]) if parts[6] else 0
                    last_num_satellites = int(parts[7]) if parts[7] else 0
                    last_hdop = float(parts[8]) if parts[8] else 0.0
                    last_altitude = float(parts[9]) if parts[9] else 0.0
                if parts[0] == '$GPGST' and parts[6] and parts[7]:
                    last_gst_timestamp = timestamp
                    last_latitude_accuracy = float(parts[6])
                    last_longitude_accuracy = float(parts[7])
                if parts[0] == '$PSSN' and parts[1] == 'HRP':
                    last_pssn_timestamp = timestamp
                    last_raw_heading = float(parts[4] or 0.0)
                    last_heading_accuracy = float(parts[7] or 'inf')
                if last_gga_timestamp == last_gst_timestamp == last_pssn_timestamp != '':
                    antenna = GeoPoint.from_degrees(last_raw_latitude, last_raw_longitude)
                    robot = antenna.polar(self._antenna_distance, -self._antenna_angle + math.radians(last_raw_heading))
                    last_latitude = math.degrees(robot.lat)
                    last_longitude = math.degrees(robot.lon)
                    last_heading = last_raw_heading - self.antenna_pose.yaw
                    self.last_measurement = GnssMeasurement(
                        time=rosys.time(),
                        pose=GeoPose.from_degrees(lat=last_latitude, lon=last_longitude, heading=last_heading),
                        latitude_std_dev=last_latitude_accuracy,
                        longitude_std_dev=last_longitude_accuracy,
                        heading_std_dev=last_heading_accuracy,
                        gps_quality=GpsQuality(last_gps_quality),
                        num_satellites=last_num_satellites,
                        hdop=last_hdop,
                        altitude=last_altitude,
                    )
                    self.NEW_MEASUREMENT.emit(self.last_measurement)
            except Exception as e:
                self.log.exception(e)

    # TODO: move to helper and add search argument; for other serial devices
    def _find_device(self) -> str:
        for port in list_ports.comports():
            self.log.debug('Found port: %s - %s', port.device, port.description)
            if 'Septentrio' in port.description:
                self.log.info('Found GNSS device: %s', port.device)
                return port.device
        raise RuntimeError('No GNSS device found')

    def _connect_to_device(self, port: str, *, baudrate: int = 115200, timeout: float = 0.2) -> serial.Serial:
        self.log.info('Connecting to GNSS device "%s"...', port)
        try:
            return serial.Serial(port=port, baudrate=baudrate, timeout=timeout)
        except serial.SerialException as e:
            raise RuntimeError(f'Could not connect to GNSS device: {port}') from e

    @staticmethod
    def _convert_to_decimal(coord: str, direction: str) -> float:
        """Convert a coordinate in the format DDMM.mmmm to decimal degrees.

        :param coord: the coordinate to convert
        :param direction: the direction (N/S/E/W)
        :return: the coordinate in decimal degrees
        """
        split_index = coord.find('.') - 2
        degrees = float(coord[:split_index])
        minutes = float(coord[split_index:])
        decimal = degrees + minutes / 60
        if direction in ['S', 'W']:
            decimal = -decimal
        return decimal


class GnssSimulation(Gnss):
    """Simulation of a GNSS receiver."""

    def __init__(self, *,
                 wheels: WheelsSimulation,
                 lat_std_dev: float = 0.01,
                 lon_std_dev: float = 0.01,
                 heading_std_dev: float = 0.01,
                 gps_quality: GpsQuality = GpsQuality.RTK_FIXED) -> None:
        """
        :param wheels: the wheels to use for the simulation
        :param lat_std_dev: the standard deviation of the latitude in meters
        :param lon_std_dev: the standard deviation of the longitude in meters
        :param heading_std_dev: the standard deviation of the heading in degrees
        :param gps_quality: the quality of the GPS signal
        """
        super().__init__()
        self.wheels = wheels
        self._is_connected = True
        self._lat_std_dev = lat_std_dev
        self._lon_std_dev = lon_std_dev
        self._heading_std_dev = heading_std_dev
        self._gps_quality = gps_quality
        self.last_measurement: GnssMeasurement | None = None
        rosys.on_repeat(self.simulate, 1.0)

    @property
    def is_connected(self) -> bool:
        return self._is_connected

    @is_connected.setter
    def is_connected(self, value: bool) -> None:
        self._is_connected = value

    def simulate(self) -> None:
        if not self.is_connected:
            return
        geo_pose = GeoPose.from_pose(self.wheels.pose)

        noise_lat = np.random.normal(0, self._lat_std_dev)
        noise_lon = np.random.normal(0, self._lon_std_dev)
        noise_heading = np.random.normal(0, math.radians(self._heading_std_dev))

        direction = math.atan2(noise_lon, noise_lat)
        noise_point = geo_pose.point.polar(noise_lat, direction)
        noise_pose = GeoPose(lat=noise_point.lat, lon=noise_point.lon, heading=geo_pose.heading + noise_heading)

        self.last_measurement = GnssMeasurement(
            time=rosys.time(),
            pose=noise_pose,
            latitude_std_dev=self._lat_std_dev,
            longitude_std_dev=self._lon_std_dev,
            heading_std_dev=self._heading_std_dev,
            gps_quality=self._gps_quality,
        )
        self.NEW_MEASUREMENT.emit(self.last_measurement)

    def developer_ui(self) -> None:
        super().developer_ui()
        ui.label('Simulation').classes('text-center text-bold')
        with ui.column().classes('gap-y-1'):
            ui.checkbox('Connected').bind_value(self, '_is_connected')
            ui.number(label='Latitude Std Dev', format='%.3f', prefix='± ', suffix='m') \
                .bind_value(self, '_lat_std_dev').classes('w-4/5')
            ui.number(label='Longitude Std Dev', format='%.3f', prefix='± ', suffix='m') \
                .bind_value(self, '_lon_std_dev').classes('w-4/5')
            ui.number(label='Heading Std Dev', format='%.2f', prefix='± ', suffix='°') \
                .bind_value(self, '_heading_std_dev').classes('w-4/5')
            ui.select({quality: quality.name for quality in GpsQuality}, label='Quality') \
                .bind_value(self, '_gps_quality').classes('w-4/5')



from ... import persistence
from ...geometry import Rectangle
from ..image_rotation import ImageRotation
from .camera import Camera


class TransformableCamera(Camera):

    def __init__(self, *, crop: Rectangle | dict | None = None, rotation: ImageRotation | int = ImageRotation.NONE, **kwargs) -> None:
        """
        A camera for which the image can be cropped and rotated.

        This mixin does not implement the actual cropping and rotation.
        It only provides the infrastructure for storing the crop and rotation parameters.

        :param crop: region to crop on the original resolution before rotation
        :param rotation: clockwise rotation which should be applied after cropping
        """
        super().__init__(**kwargs)

        self.crop: Rectangle | None = None
        if isinstance(crop, Rectangle):
            self.crop = crop
        elif isinstance(crop, dict):
            self.crop = persistence.from_dict(Rectangle, crop)

        self.rotation: ImageRotation = rotation if isinstance(
            rotation, ImageRotation) else ImageRotation.from_degrees(rotation)

    def to_dict(self) -> dict:
        return super().to_dict() | {
            'crop': persistence.to_dict(self.crop),
            'rotation': self.rotation.value,
        }

    @property
    def rotation_angle(self) -> int:
        """Clockwise rotation angle in degrees."""
        return int(self.rotation)

    @rotation_angle.setter
    def rotation_angle(self, value: int) -> None:
        self.rotation = ImageRotation.from_degrees(value)

    def rotate_clockwise(self) -> None:
        """Rotate the image clockwise by 90 degrees."""
        self.rotation_angle += 90

    def rotate_counter_clockwise(self) -> None:
        """Rotate the image counter clockwise by 90 degrees."""
        self.rotation_angle -= 90



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from nicegui import ui
from nicegui.elements.mixins.value_element import ValueElement

if TYPE_CHECKING:
    from .robot_brain import RobotBrain


class StatusBulb(ValueElement):
    def __init__(self, value: bool = False) -> None:
        super().__init__(value=value, on_value_change=self.on_change, tag='span')
        self.style('height: 15px; width: 15px; margin: auto; border-radius: 50%')
        self.on_change()

    def on_change(self) -> None:
        self.style('background: radial-gradient(circle at 5px 5px, #5898D4, #4682B4);' if self.value
                   else 'background: radial-gradient(circle at 5px 5px, #D3D3D3, #A9A9A9);')


@dataclass(slots=True, kw_only=True)
class GpioPin:
    gpio: int
    level: bool = False
    is_input: bool = False
    is_output: bool = False
    open_drain: bool = False
    is_pullup: bool = False
    is_pulldown: bool = False
    drive_strength: int = 0
    sleep_sel: int = 0


class EspPins:
    """Monitor and control ESP32 GPIO pins."""

    def __init__(self, name: str, robot_brain: RobotBrain) -> None:
        self.name = name
        self.robot_brain = robot_brain
        _pin_numbers: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                                   10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                                   21, 22, 23, 25, 26, 27,
                                   32, 33, 34, 35, 36, 37, 38, 39]
        self._pins: dict[int, GpioPin] = {pin: GpioPin(gpio=pin) for pin in _pin_numbers}

    async def update_pin(self, pin: GpioPin) -> None:
        ack = f'GPIO_Status[{pin.gpio}]|' if self.name == 'core' else f'{self.name}:'
        line = await self.robot_brain.send_and_await(f'{self.name}.get_pin_status({pin.gpio})', ack, timeout=1.0)
        if line is None:
            raise TimeoutError(f'No response from {self.name} for pin {pin.gpio}')

        # GPIO_Status[3]| Level: 1| InputEn: 1| OutputEn: 0| OpenDrain: 0| Pullup: 1| Pulldown: 0| DriveStrength: 2| SleepSel: 0
        # p0: GPIO_Status[0]| Level: 1| InputEn: 1| OutputEn: 0| OpenDrain: 0| Pullup: 1| Pulldown: 0| DriveStrength: 2| SleepSel: 0
        status_dict = {}
        for part in line.split('|')[1:]:
            key, value = part.strip().split(': ')
            status_dict[key.strip()] = int(value.strip())

        pin.level = bool(status_dict.get('Level', 0))
        pin.is_input = bool(status_dict.get('InputEn', 0))
        pin.is_output = bool(status_dict.get('OutputEn', 0))
        pin.open_drain = bool(status_dict.get('OpenDrain', 0))
        pin.is_pullup = bool(status_dict.get('Pullup', 0))
        pin.is_pulldown = bool(status_dict.get('Pulldown', 0))
        pin.drive_strength = int(status_dict.get('DriveStrength', 0))
        pin.sleep_sel = int(status_dict.get('SleepSel', 0))

    async def update_all(self) -> None:
        for pin in self._pins.values():
            await self.update_pin(pin)

    async def set_pin_level(self, pin: GpioPin, level: bool) -> None:
        await self.robot_brain.send(f'{self.name}.set_pin_level({pin.gpio}, {1 if level else 0})')

    def developer_ui(self) -> None:
        with ui.column():
            with ui.row().classes('w-full'):
                ui.markdown(f'**ESP: {self.name}**')
                ui.space()
                ui.button(icon='refresh', on_click=self.update_all).props('flat round dense').tooltip('Update once')
            ui.separator()
            with ui.grid(columns=5):
                for pin in self._pins.values():
                    with ui.row():
                        ui.label(f'GPIO {str(pin.gpio).zfill(2)}')
                        with StatusBulb().bind_value_from(pin, 'level'):
                            self._pin_context_menu(pin)

    def _pin_context_menu(self, pin: GpioPin) -> None:
        with ui.context_menu().props('auto-close=false') as context_menu:
            with ui.card():
                with ui.row().classes('w-full'):
                    ui.label(f'GPIO {pin.gpio}')
                    ui.space()
                    ui.button(icon='refresh', on_click=lambda pin=pin: self.update_pin(pin)) \
                        .props('size=sm flat dense').tooltip('Update gpio state')
                    ui.button(icon='check_circle', on_click=lambda pin=pin: self.set_pin_level(pin, True)) \
                        .props('size=sm flat dense').tooltip('Set output to HIGH')
                    ui.button(icon='unpublished', on_click=lambda pin=pin: self.set_pin_level(pin, False)) \
                        .props('size=sm flat dense').tooltip('Set output to LOW')
                    ui.space()
                    ui.button(icon='close', on_click=context_menu.close) \
                        .props('color=black size=sm flat dense')
                ui.separator()
                with ui.grid(columns=2):
                    ui.label().bind_text_from(pin, 'level', lambda level: f'Level: {"High" if level else "Low"}')
                    ui.label().bind_text_from(pin, 'is_input', lambda is_input: f'Input: {is_input}')
                    ui.label().bind_text_from(pin, 'is_output', lambda is_output: f'Output: {is_output}')
                    ui.label().bind_text_from(pin, 'open_drain', lambda open_drain: f'OpenDrain: {open_drain}')
                    ui.label().bind_text_from(pin, 'is_pullup', lambda is_pullup: f'PullUp: {is_pullup}')
                    ui.label().bind_text_from(pin, 'is_pulldown', lambda is_pulldown: f'PullDown: {is_pulldown}')
                    ui.label().bind_text_from(pin, 'drive_strength', lambda strength: f'Drive Strength: {strength}')
                    ui.label().bind_text_from(pin, 'sleep_sel', lambda sleep_sel: f'Sleep Sel: {sleep_sel}')
